import ipyvuetify as v
from .figure_card import FigureCard
import plotly.express as px


# Table Header Class
class TableHeader(v.Card):
    def __init__(self, *args, **kwargs):
        class_ = 'd-flex justify-space-between'
        dense = True
        outlined = False
        elevation = 0
        super().__init__(class_=class_,
                         dense=dense,
                         outlined=outlined,
                         elevation=elevation,
                         *args, **kwargs)

        self.checkbox_header = v.SimpleCheckbox(v_model=False, class_='d-flex justify-center', width='10%',
                                                style_='overflow: hidden')
        self.name_header = v.Card(children=['Name'], background='none', elevation=0, width='55%', class_='px-3 py-1',
                                  style_='font-weight:bold')
        self.unit_header = v.Card(children=['Unit'], background='none', elevation=0, width='15%',
                                  class_='d-flex justify-center px-3 py-1', style_='font-weight:bold')
        self.color_header = v.Card(children=['Color'], background='none', elevation=0, width='15%',
                                   class_='d-flex justify-center py-1', style_='font-weight:bold')
        self.line_style_header = v.Card(children=['Line Style'], background='none', elevation=0, width='20%',
                                        class_='d-flex justify-center py-1 mx-2 px-2', style_='font-weight:bold')

        self.children = [
            v.Card(children=[self.checkbox_header], flat=True, elevation=0, width='7%', class_='d-flex justify-center'),
            v.Divider(vertical=True),
            self.name_header,
            v.Divider(vertical=True),
            self.unit_header,
            v.Divider(vertical=True),
            self.color_header,
            v.Divider(vertical=True),
            self.line_style_header]


# Table Item Class (Row Item)
class TableItem(v.Card):
    item_id = 0
    colors = px.colors.qualitative.Plotly

    def __init__(self, name=None, unit='-', *args, **kwargs):
        rc = ['grey lighten-4', 'white']

        if not name:
            name = 'Data ' + str(TableItem.item_id + 1)
        class_ = 'd-flex justify-space-between'
        color = rc[TableItem.item_id % 2]
        dense = True
        outlined = False
        elevation = 0

        super().__init__(class_=class_,
                         dense=dense,
                         outlined=outlined,
                         elevation=elevation,
                         color=color,
                         *args, **kwargs)

        self.name_item = v.Card(children=[name],
                                color=rc[TableItem.item_id % 2],
                                elevation=0,
                                width='55%',
                                class_='px-3 py-1')

        self.unit_item = v.Card(children=[unit],
                                color=rc[TableItem.item_id % 2],
                                elevation=0,
                                width='15%',
                                class_='d-flex justify-center px-3 py-1',
                                align_center=True)

        self.color_item = v.Card(children=[v.Icon(children=['mdi-checkbox-blank'],
                                                  color=TableItem.colors[TableItem.item_id % 10],
                                                  class_='',
                                                  dense=True)],
                                 color=rc[TableItem.item_id % 2],
                                 elevation=0,
                                 width='15%',
                                 class_='d-flex justify-center py-1')

        self.style_item = Line_Style(idx=TableItem.item_id,
                                     style_='font-weight:bold; height:26px; width:20%; overflow: hidden hidden'
                                     )

        self.checkbox_item = v.SimpleCheckbox(v_model=False, class_='d-flex justify-left', width='10%',
                                              style_='overflow: hidden', model_id='{}'.format(TableItem.item_id))

        self.children = [v.Card(children=[self.checkbox_item], elevation=0, width='7%', class_='d-flex justify-center',
                                color=rc[TableItem.item_id % 2]),
                         v.Divider(vertical=True),
                         self.name_item,
                         v.Divider(vertical=True),
                         self.unit_item,
                         v.Divider(vertical=True),
                         self.color_item,
                         v.Divider(vertical=True),
                         self.style_item
                         ]

        TableItem.item_id += 1

    # Reset Tag ID
    @staticmethod
    def reset():
        TableItem.item_id = 0


# Table Class (TableHeader + TableItem)
class TableCard(v.Card):
    def __init__(self, *args, **kwargs):
        TableItem().reset()
        width = '60%'
        flat = False
        elevation = 5
        class_ = 'pa-4 ma-4'
        style_ = 'border-radius:12px; overflow:hidden auto; height:30%'
        super().__init__(width=width,
                         flat=flat,
                         class_=class_,
                         style_=style_,
                         elevation=elevation,
                         *args, **kwargs)

        self.table_header = TableHeader()
        self.items_list = []
        self.empty_table = v.Row(children=['No Data Available'], class_='d-flex justify-center pa-1 pt-3')

        self.children = [self.table_header, v.Divider(), self.empty_table]

    def create_table(self, items_list):
        self.items_list = items_list
        table = [self.table_header, v.Divider()] + items_list
        self.children = table


class Line_Style(v.Select):
    def __init__(self, idx=None, style_=None, *args, **kwargs):
        super().__init__(style_=style_,
                         *args, **kwargs)

        self.items = ['____', '_ _ _', '.....', '_ . _']
        self.v_model = self.items[0]

        self.style_dict = {'____': 'solid', '_ _ _': 'dash', '_ . _': 'dashdot', '.....': 'dot'}
        self.dense = True
        self.class_ = 'px-2 mx-2'
        self.idx = idx

    def get_style(self):
        return self.style_dict[self.v_model]



from copy import deepcopy

import ipyvuetify as v
from pandas import DataFrame

from ._backend import push_formula, push_signal

from .figure_table import FigureTable
from .panels import LeftPanel, ARXPanel, SSPanel, NNPanel

from seeq_sysid.model.base import Model
from seeq_sysid.model.arx import ARX
from seeq_sysid.model.ss import Subspace
from seeq_sysid.model.nn import NN


class AppSheet(v.Card):
    def __init__(self,
                 panel: LeftPanel,
                 workbook_id='',
                 *args, **kwargs):
        class_ = 'd-flex justify-space-between ma-2 pa-2 pt-2 mt-0'
        style_ = 'height:900px; border-radius:12px'
        color = 'white'
        flat = True
        super().__init__(class_=class_,
                         style_=style_,
                         color=color,
                         flat=flat,
                         elevation=0,
                         **kwargs)

        self.train_results = None
        self.validation_results = None
        self.all_results = None

        self.panel = panel
        self.canvas = FigureTable()

        self.model_name = panel.model_name
        self.model = Model()
        self.blank_model = deepcopy(self.model)
        self.addon_worksheet = 'From ' + self.model_name + ' Addon'

        # General Widgets
        self.mv_select = panel.mv_select
        self.cv_select = panel.cv_select

        self.train_condition = panel.train_condition
        self.validation_condition = panel.validation_condition

        self.push_model_btn = panel.push_model_btn
        self.identify_model_btn = panel.identify_model_btn
        self.validate_model_btn = panel.validate_model_btn

        self.signal_df = DataFrame()
        self.capsule_df = DataFrame()
        self.tags_df = DataFrame()
        self.workbook_id = workbook_id

        # Actions
        self.mv_select.on_event('change', self.update_panel)
        self.cv_select.on_event('change', self.update_panel)
        self.identify_model_btn.on_event('click', self.identify_system)
        self.validate_model_btn.on_event('click', self.validate_model)
        self.push_model_btn.on_event('click', self.push_model)

        self.children = [self.panel, self.canvas]

    # Functions
    def set_data(self, signal_df: DataFrame = DataFrame(), capsule_df: DataFrame = DataFrame(), tags_df=DataFrame(),
                 workbook_id: str = ''):
        self.mv_select.items = signal_df.columns.to_list()
        self.cv_select.items = signal_df.columns.to_list()

        self.train_condition.items = capsule_df.columns.to_list()
        self.validation_condition.items = capsule_df.columns.to_list()

        self.signal_df = signal_df
        self.capsule_df = capsule_df
        self.tags_df = tags_df
        self.workbook_id = workbook_id

    def update_panel(self, *_):
        self.mv_select.items = [item for item in self.signal_df.columns if item not in self.cv_select.v_model]
        self.cv_select.items = [item for item in self.signal_df.columns if item not in self.mv_select.v_model]

        if len(self.mv_select.v_model) > 0 and len(self.cv_select.v_model) > 0:
            self.identify_model_btn.disabled = False

        else:
            self.identify_model_btn.disabled = True
            self.validate_model_btn.disabled = True
            self.push_model_btn.disabled = True

    def identify_system(self, *_):
        self.identify_model_btn.loading = True
        self.prepare_params_general()
        self.prepare_params_spec()

        train_dataset = self.create_dataset(self.train_condition.v_model)
        validation_dataset = self.create_dataset(self.validation_condition.v_model)

        self.preprocess()

        self.model.identify(train_dataset)

        self.train_results = self.model.forecast(train_dataset)
        self.train_results.set_index(train_dataset.index, inplace=True)
        self.train_results[self.model.cv] = train_dataset[self.model.cv]

        self.validation_results = self.model.forecast(validation_dataset)
        self.validation_results.set_index(validation_dataset.index, inplace=True)
        self.validation_results[self.model.cv] = validation_dataset[self.model.cv]

        self.canvas.create(self.train_results, self.validation_results)

        if self.model.status:
            self.push_model_btn.disabled = False
            self.validate_model_btn.disabled = False
        else:
            self.push_model_btn.disabled = True
            self.validate_model_btn.disabled = True

        self.identify_model_btn.loading = False

    def create_dataset(self, capsules: list = None):
        signal_df = self.signal_df[self.mv_select.v_model + self.cv_select.v_model]
        capsule_df = self.capsule_df

        if not capsules:
            return signal_df

        return signal_df[capsule_df[capsules].sum(axis=1) == True]

    def validate_model(self, *_):
        self.validate_model_btn.loading = True
        validation_dataset = self.create_dataset(self.validation_condition.v_model)
        self.validation_results = self.model.forecast(validation_dataset)
        self.validation_results.set_index(validation_dataset.index, inplace=True)
        self.validation_results[self.model.cv] = validation_dataset[self.model.cv]

        self.canvas.create(self.train_results, self.validation_results)
        self.validate_model_btn.loading = False

    def push_model(self, *_):
        if self.general_validation():
            return None

        self.model.create_formula(self.tags_df)
        # Measured Data
        signal_df = self.signal_df[self.model.cv]
        push_formula(signal_df, self.model.formula, self.workbook_id, self.addon_worksheet)

    def push_data(self, *_):
        signal_df = self.create_dataset()
        self.all_results = self.model.forecast(signal_df)
        self.all_results.set_index(signal_df.index, inplace=True)
        self.all_results[self.model.cv] = signal_df[self.model.cv]

        push_signal(self.all_results, self.workbook_id, self.addon_worksheet)

    def general_validation(self):
        return 0

    def customized_validation(self):
        return 0

    def prepare_params_general(self):
        self.model = deepcopy(self.blank_model)

        self.model.mv = self.mv_select.v_model
        self.model.cv = self.cv_select.v_model

    def prepare_params_spec(self):
        pass

    def preprocess(self):
        pass


class ARXAppSheet(AppSheet):
    def __init__(self,
                 *args, **kwargs):
        panel = ARXPanel()
        super().__init__(panel=panel,
                         *args, **kwargs)
        self.model = ARX()
        self.blank_model = deepcopy(self.model)

        self.model_struct = self.panel.model_struct_select

        self.na_min = self.panel.na_min
        self.na_max = self.panel.na_max

        self.nb_min = self.panel.nb_min
        self.nb_max = self.panel.nb_max

        self.nk_min = self.panel.nk_min
        self.nk_max = self.panel.nk_max

    def prepare_params_spec(self):
        self.model.na_min = int(self.na_min.v_model)
        self.model.na_max = int(self.na_max.v_model)

        self.model.nb_min = int(self.nb_min.v_model)
        self.model.nb_max = int(self.nb_max.v_model)

        self.model.nk_min = int(self.nk_min.v_model)
        self.model.nk_max = int(self.nk_max.v_model)

        self.model.model_struct = self.model_struct.v_model


class SSAppSheet(AppSheet):
    def __init__(self,
                 *args, **kwargs):
        panel = SSPanel()
        super().__init__(panel=panel,
                         *args, **kwargs)
        self.model = Subspace()
        self.blank_model = deepcopy(self.model)

        self.method = self.panel.method_select
        self.thresh = self.panel.threshold_box
        self.order = self.panel.order_box
        self.multiplier_min = self.panel.multiplier_min
        self.multiplier_max = self.panel.multiplier_max

        self.push_model_btn.on_event('click', self.push_data)

    def prepare_params_spec(self):
        self.model.method = self.method.v_model
        self.model.thresh = float(self.thresh.v_model)
        self.model.order = int(self.order.v_model)
        self.model.om_min = int(self.multiplier_min.v_model)
        self.model.om_max = int(self.multiplier_max.v_model)


class NNAppSheet(AppSheet):
    def __init__(self,
                 *args, **kwargs):
        panel = NNPanel()
        super().__init__(panel=panel,
                         *args, **kwargs)
        self.model = NN()
        self.blank_model = deepcopy(self.model)

        self.model.mode = int(self.panel.auto_mode_slider.v_model)

        self.push_model_btn.on_event('click', self.push_data)

    def prepare_params_spec(self):
        self.model.mode = int(self.panel.auto_mode_slider.v_model)

    def preprocess(self):
        self.model.Min = self.signal_df[self.model.mv + self.model.cv].min()
        self.model.Max = self.signal_df[self.model.mv + self.model.cv].max()



from copy import deepcopy

import ipyvuetify as v
import plotly.express as px
import plotly.graph_objects as go
from pandas import DataFrame


# Figures Tabs (Train+Validation) Class
class FigureCard(v.Card):
    def __init__(self,
                 width='60%',
                 height='auto',
                 elevation=10,
                 shaped=True,
                 outlined=False,
                 color='white',
                 class_='ma-4 pa-4',
                 **kwargs):

        super().__init__(width=width,
                         height=height,
                         elevation=elevation,
                         shaped=shaped,
                         outlined=outlined,
                         color=color,
                         class_=class_,
                         **kwargs)

        fig = go.FigureWidget()

        fig['layout']['showlegend'] = False
        fig['layout']['xaxis']['rangeslider']['visible'] = True
        fig['layout']['xaxis']['title'] = 'Time'
        fig['layout']['margin']['t'] = 30
        fig['layout']['margin']['b'] = 40
        fig['layout']['margin']['r'] = 50
        fig['layout']['yaxis']['title'] = 'Value'
        fig['layout']['yaxis']['fixedrange'] = False

        self.colors = px.colors.qualitative.Plotly

        self.blank_fig = fig
        self.train_fig = deepcopy(self.blank_fig)
        self.validation_fig = deepcopy(self.blank_fig)

        self.create_figure()

    def create_figure(self):
        train_tab = v.Tab(children=['Train'], style_='font-weight:bold')
        validation_tab = v.Tab(children=['Validation'], style_='font-weight:bold')

        train_item = v.TabItem(children=[self.train_fig])
        validation_item = v.TabItem(children=[self.validation_fig])

        figure_tabs = v.Tabs(children=[train_tab, validation_tab, train_item, validation_item], fixed_tabs=True,
                             color='#1d376c')
        self.children = [figure_tabs]

    def plot(self, train_df: DataFrame, validation_df: DataFrame):
        self.train_fig = deepcopy(self.blank_fig)
        self.validation_fig = deepcopy(self.blank_fig)

        tag_id = 0
        for tag in train_df.columns:
            self.train_fig.add_trace(go.Scatter(x=train_df.index, y=train_df[tag], name=tag, visible=False,
                                                line=dict(color=self.colors[tag_id % 10])))
            tag_id += 1

        tag_id = 0
        for tag in validation_df.columns:
            self.validation_fig.add_trace(
                go.Scatter(x=validation_df.index, y=validation_df[tag], name=tag, visible=False,
                           line=dict(color=self.colors[tag_id % 10])))
            tag_id += 1

        self.create_figure()

    def update(self, idx: int, status: bool = None, line_style: str = None):
        if status is not None:
            self.train_fig.data[idx].visible = status
            self.validation_fig.data[idx].visible = status

        if line_style:
            self.train_fig.data[idx].line['dash'] = line_style
            self.validation_fig.data[idx].line['dash'] = line_style



from gekko import GEKKO

from numpy import vstack, zeros, reshape, array
from pandas import DataFrame

from .base import Model
from re import split


def create_formula_variable_name(names):
    formula_name = []
    for item in names:
        item = item.lower()
        item = split("_| | ", item)
        f_name = '$'
        for i in item:
            f_name += i
        formula_name.append(f_name)

    return formula_name


class ARX(Model):
    """ AutoRegressive with Exogenous input (ARX)
    Model Structure:
    y(k+1) = a_0*y(k)+a_1*y(k-1)+...+a_(na-1)*y(k-na+1)
              + b_0+u(k)+b_1*u(k-1)+...+b_(nb-1)*u(k-na-nk+1)
    where:
        u: Input(s)   [pandas DataFrame]
        y: Output(s)  [pandas DataFrame]
        na: AutoRegressive term order
            na=2 -> a_0*y(k) + a_1*y(k-1)
        nb: Exogenous input order
            nb=2 -> b_0*u(k) + b_1*u(k-1)
        nk: Input Delay
            nk=2 -> b_0*u(k-2) + b_1*u(k-3)
        n_u: Number of Inputs
        n_y: Number of Outputs
    """

    def __init__(self):

        super().__init__()
        self.model_struct = 'ARX'

        self.p = None
        # AR part order
        self.na_min = 0
        self.na_max = 2
        self.na = None
        # Exogenous part order
        self.nb_min = 0
        self.nb_max = 2
        self.nb = None
        # Input delay
        self.nk_min = 0
        self.nk_max = 0
        self.nk = None

    def identify(self, df: DataFrame = None):

        self.df = df
        u_df = df[self.mv]
        y_df = df[self.cv]

        t = list(range(len(y_df.index)))

        for i_a in range(self.na_min, self.na_max + 1):
            for i_b in range(self.nb_min, self.nb_max + 1):
                for i_k in range(self.nk_min, self.nk_max + 1):
                    model = GEKKO(remote=False)
                    yp, p, _K = model.sysid(t=t, u=u_df, y=y_df, na=i_a, nb=i_b, nk=i_k, pred='meas', shift='calc')

                    e = (y_df.to_numpy() - yp) ** 2
                    e = sum(sum(e)) / len(e)

                    if e < self.error_best:
                        self.error_best = e
                        self.p = p
                        self.na = i_a
                        self.nb = i_b
                        self.nk = i_k
                        self.yp = yp

        self.status = True

    def predict(self, df: DataFrame = None):
        u_df = df[self.mv]
        y_df = df[self.cv]

        ly, ny = y_df.shape
        lu, nu = u_df.shape
        na = self.na
        nb = self.nb
        nk = self.nk

        alpha = reshape(self.p['a'].transpose(), [ny, na])
        beta = self.p['b']
        c = self.p['c']

        nd_max = max(na, nb + nk)

        yp: array = y_df.iloc[:nd_max].to_numpy()
        ni = max(0, na - nb - nk)

        if self.model_struct == 'ARX':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = alpha[i][::-1].dot(y_df[k:k + na].iloc[:, [i]]) + c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])

                yp = vstack([yp, y_p])

        elif self.model_struct == 'FIR':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    # y_p[i] = C[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])

                yp = vstack([yp, y_p])

        labels = []
        for label in y_df.columns:
            labels.append(label + '_pred')

        yp = DataFrame(yp, columns=labels)
        yp.fillna(method='ffill', inplace=True)
        yp.fillna(method='bfill', inplace=True)
        self.yp = yp

        return yp

    def forecast(self, df: DataFrame = None):
        u_df = df[self.mv]
        y_df = df[self.cv]

        ly, ny = y_df.shape
        lu, nu = u_df.shape
        na = self.na
        nb = self.nb
        nk = self.nk

        alpha = reshape(self.p['a'].transpose(), [ny, na])
        beta = self.p['b']
        c = self.p['c']

        nd_max = max(na, nb + nk)

        yp = y_df.iloc[:nd_max].to_numpy()
        yo = yp[-na:].transpose()

        ni = max(0, na - nb - nk)

        if self.model_struct == 'ARX':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = alpha[i][::-1].dot(yo[i]) + c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])
                    yo[i][:na - 1] = yo[i][1:]
                    yo[i][-1] = y_p[i]
                yp = vstack([yp, y_p])

        elif self.model_struct == 'FIR':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])
                    yo[i][:na - 1] = yo[i][1:]
                    yo[i][-1] = y_p[i]
                yp = vstack([yp, y_p])

        self.label = [tag_label + '_pred' for tag_label in y_df.columns]

        yp_df = DataFrame(yp, columns=self.label)
        yp_df.fillna(method='ffill', inplace=True)
        yp_df.fillna(method='bfill', inplace=True)
        return yp_df

    def create_formula(self, tags):
        u_name = self.mv
        y_name = self.cv

        n_y = self.p['a'].shape[1]
        n_u = self.p['b'].shape[2]
        n_a = self.p['a'].shape[0]
        n_bk = self.p['b'].shape[1]

        yf_name = create_formula_variable_name(y_name)
        uf_name = create_formula_variable_name(u_name)

        formula_dic = {}
        for tag in range(len(yf_name)):
            formula_dic[yf_name[tag]] = tags[tags['Name'] == y_name[tag]]

        for tag in range(len(uf_name)):
            formula_dic[uf_name[tag]] = tags[tags['Name'] == u_name[tag]]

        time_interval = self.df.index[1] - self.df.index[0]

        timestep: float = time_interval.total_seconds()

        formula_list = []

        for i in range(n_y):
            formula = ''
            for j in range(n_a):
                formula += ' {}.move({}s)*({})\n+'.format(yf_name[i], (j + 1) * (-timestep), self.p['a'][j][i])

            for k in range(n_u):
                for j in range(n_bk):
                    formula += ' {}.move({}s)*({})\n+'.format(uf_name[k], (j + 1) * (-timestep), self.p['b'][i][j][k])

            formula += str(self.p['c'][i])

            formula_list.append({
                'Name': '{} arx'.format(y_name[i]),
                'Type': 'CalculatedSignal',
                'Description': 'ARX Model {} Formula '.format(y_name[i]),
                'Formula': formula,
                'Formula Parameters': formula_dic
            })

        self.formula = DataFrame(formula_list)

# from pandas import read_csv
# import matplotlib.pyplot as plt
#
#
# x_df = read_csv('../../signal_df.csv')
# x_df.set_index('Time', drop=True, inplace=True)
# x_df = x_df.iloc[50:]
# arx = ARX()
#
# arx.mv = ['F_cw']
# arx.cv = ['T1', 'T2', ]
# arx.identify(x_df)
# yp = arx.forecast(x_df)
# plt.plot(yp)
# plt.show()



from .app_sheet import AppSheet, ARXAppSheet, SSAppSheet, NNAppSheet
from .app_bar import AppBar
from .panels import LeftPanel, ARXPanel, SSPanel, NNPanel
from .figure_table import FigureTable
from .utils import *



from pandas import DataFrame


class Model:
    def __init__(self):
        self.status = False
        self.label = None
        self.model = None
        self.df = None
        self.label = None
        self.formula = None
        self.mv = None
        self.cv = None
        self.yp = None
        self.error = None
        self.error_best = 1e9
        self.eval_best = 1e9
        self.thresh = 1e-6
        self.order = 4

        self.u_ss = None
        self.x_ss = None
        self.y_ss = None
        # self.n = None
        # self.m = None
        # self.d = None
        #
        # self.p = None
        # self.q = None

        # self.eps = None
        # self.A = None
        # self.B = None
        # self.C = None
        # self.D = None

        # self.alpha = None
        # self.beta = None

    def identify(self, df: DataFrame = None):
        pass

    def predict(self, df: DataFrame = None):
        pass

    def forecast(self, df: DataFrame = None):
        pass



from pandas import DataFrame
from math import floor


def test_train_split(x_df: DataFrame = DataFrame(), y_df: DataFrame = DataFrame(), train_ratio: float = 0.75) \
        -> (DataFrame, DataFrame, DataFrame, DataFrame):
    if x_df.empty or y_df.empty:
        return x_df, x_df, y_df, y_df

    k = len(x_df)

    x_train: DataFrame = x_df.iloc[:floor(k * train_ratio)]
    x_valid: DataFrame = x_df.iloc[floor(k * train_ratio):]
    y_train: DataFrame = y_df.iloc[:floor(k * train_ratio)]
    y_valid: DataFrame = y_df.iloc[floor(k * train_ratio):]

    return x_train, x_valid, y_train, y_valid


def shifter(df_o: DataFrame, dummy_state) -> DataFrame:
    df_n = DataFrame()
    for i in range(dummy_state):
        for tag in df_o.columns:
            df_n[tag + '_d{}'.format(i)] = df_o[tag].shift(i)
    df_n.fillna(method='bfill', inplace=True)

    return df_n



# Custom
from ipywidgets import HTMLMath, Layout
from ipyvuetify import Tooltip


def create_eq(text, color, fontsize, top='10px', bot=''):
    """Create a latex equation in the gui"""
    return HTMLMath(value=f"<b><font color={color}><font size={fontsize}>{text}</b>",
                    layout=Layout(top=top, bottom=bot))


def add_tooltip(item, text):
    """Create a tooltip for the input ipyvuetify widget"""
    item.v_on = 'tooltip.on'
    return Tooltip(bottom=True, v_slots=[{
        'name': 'activator',
        'variable': 'tooltip',
        'children': item,
    }], children=[text])



import ipyvuetify as v
import ipywidgets as widgets
from pathlib import Path
from .utils import add_tooltip


class HamburgerMenu(v.Menu):
    """
    Create a Hamburger Menu in the right corner of App Bar
    """

    def __init__(self, **kwargs):
        # Create menu icon button
        self.hamburger_button = v.AppBarNavIcon(v_on='menuData.on', class_='align-center ma-2 mt-3')
        # Feedback option
        self.feedback_button = v.ListItem(value='help',
                                          ripple=True,
                                          href='https://github.com/BYU-PRISM/Seeq/issues',
                                          children=[v.ListItemAction(class_='mr-2 ml-0',
                                                                     children=[v.Icon(color='#212529',
                                                                                      children=['fa-life-ring'])]),
                                                    v.ListItemActionText(children=[f'Send Support Request'])
                                                    ])
        # GitHub page
        self.user_guide_button = v.ListItem(value='tutorial',
                                            ripple=True,
                                            href='https://github.com/BYU-PRISM/Seeq',
                                            target='_blank',
                                            children=[v.ListItemAction(class_='mr-2 ml-0',
                                                                       children=[v.Icon(color='#212529',
                                                                                        children=['mdi-help-box'])]),
                                                      v.ListItemActionText(children=[f'User Guide'])
                                                      ])

        self.items = [v.Divider(), self.feedback_button, v.Divider(), self.user_guide_button, v.Divider()]

        super().__init__(offset_y=True,
                         offset_x=False,
                         dense=True,
                         left=True,
                         v_slots=[{
                             'name': 'activator',
                             'variable': 'menuData',
                             'children': self.hamburger_button
                         }],
                         children=[
                             v.List(children=self.items)
                         ],
                         **kwargs)


class AppBar(v.Card):
    """
    Create app bar for the Add-on. This app bar includes icon, tabs, load worksheet button and, hamburger menu.
    """

    def __init__(self, tabs=None, items=None, logo_name=Path(__file__).parent / '../data/seeq_logo.png', *args,
                 **kwargs):
        """
        Args:
            tabs : list of tabs
            items : list of items (app sheet)
            logo_name : App icon directory
        """
        super().__init__(*args, **kwargs)
        if items is None:
            items = []
        if tabs is None:
            tabs = []
        # read add-on logo (.png)
        file = open(logo_name, 'rb')
        image = file.read()
        self.image = widgets.Image(value=image, format='png', align='center')
        file.close()

        self.logo = [v.Img(children=[self.image], class_='align-center mt-2 pl-4')]
        # load worksheet button
        self.load_worksheet = v.Btn(icon=True, children=[v.Icon(children=['mdi-folder-search-outline'])],
                                    class_='align-center ma-2 mt-3')
        self.load_worksheet.on_event('click', self.load_worksheet_action)

        self.close_url_dialog_btn = v.Btn(children=['CLOSE'], color='#007960', text=True)
        self.close_url_dialog_btn.on_event('click', self.close_url_action)

        self.ok_url_dialog_btn = v.Btn(children=['OK'], color='#007960', text=True, loading=False)
        self.ok_url_dialog_btn.on_event('click', self.ok_url_action)
        # Create a text field for the worksheet url
        self.worksheet_url = v.TextField(v_model='', placeholder='Worksheet URL', class_='mx-4', color='#007960',
                                         clearable=True)
        self.worksheet_url.on_event('paste.stop', lambda *args: None)

        control_dialog_btn_layout = v.Layout(children=[v.Spacer(), self.close_url_dialog_btn, self.ok_url_dialog_btn])

        dialog_card_content = [v.CardTitle(children=['Please Enter a Worksheet URL:']),
                               self.worksheet_url,
                               control_dialog_btn_layout]

        dialog_card = v.Card(children=dialog_card_content, class_='pa-2 ma-3 my-0', flat=True)

        self.url_dialog = v.Dialog(name='OpneWB',
                                   children=[v.Card(children=[dialog_card])],
                                   v_model=False,
                                   max_width='600px')

        self.url_dialog.on_event('keydown.stop', lambda *args: None)
        # Create an instance of hamburger menu
        self.ham_menu = HamburgerMenu()

        # Adding Tooltips (tt = with tooltip)
        self.load_worksheet_tt = add_tooltip(self.load_worksheet, 'Open Worksheet')

        app = v.Tabs(children=self.logo + tabs + items + [v.Spacer(),
                                                          self.url_dialog,
                                                          self.load_worksheet_tt,
                                                          v.Divider(vertical=True, inset=True),
                                                          self.ham_menu],
                     color='#1d376c',
                     class_='align-center pt-0',
                     slider_size='4',
                     background_color='grey lighten-4',
                     hide_slider=True,
                     height=60,
                     style_='width:98%; border-radius:12px')

        self.children = [app]

    # Load worksheet event functions
    def load_worksheet_action(self, *args):
        self.url_dialog.v_model = True

    def close_url_action(self, *args):
        self.url_dialog.v_model = None

    def ok_url_action(self, *args):
        self.url_dialog.v_model = None



from numpy import hstack, zeros, linalg
from pandas import DataFrame

from .base import Model
from .utils import test_train_split, shifter


# Subspace Model
class Subspace(Model):
    def __init__(self):
        super().__init__()
        self.method = 'Least Square'
        
        # Order Multiplier
        self.om_min = None
        self.om_max = None

    def identify(self,
                 df: DataFrame = None):

        u_df = df[self.mv]
        y_df = df[self.cv]

        self.u_ss = u_df.head(1).to_numpy()
        self.y_ss = y_df.head(1).to_numpy()

        u_df -= self.u_ss
        y_df -= self.y_ss

        self.label = [tag_label + '_subspace' for tag_label in y_df.columns]

        u_train, u_valid, y_train, y_valid = test_train_split(u_df, y_df)

        n_step, m = u_valid.shape

        eval_0 = 1e9
        eval_best = eval_0
        order_penalty = 1e-5
        order = 1

        if self.method == 'Least Square':
            u_valid = u_valid.to_numpy()
            y_valid = y_valid.to_numpy()
            for r in range(self.om_min, self.om_max+1):
                x_df = shifter(y_train, r)

                n_x = len(x_df.columns)
                n_y = len(y_df.columns)

                yp = x_df.shift(1)
                yp = yp.fillna(method='bfill').to_numpy()
                x_train = x_df.to_numpy()

                xp = hstack([yp, u_train])

                G = linalg.lstsq(xp, x_train, rcond=None)[0].T

                A = G[:, :-m]
                B = G[:, -m:]
                C = zeros((n_y, A.shape[0]))
                for i in range(n_y):
                    C[i, i] = 1

                D = zeros((n_y, m))

                xh = zeros((n_step, n_x))
                yh = zeros((n_step, n_y))

                for i in range(1, n_step):
                    xh[i] = A.dot(xh[i - 1]) + B.dot(u_valid[i - 1])
                    yh[i] = C.dot(xh[i - 1])

                e_val = ((yh - y_valid) ** 2).sum() / len(yh) ** 2

                if e_val + (A.shape[0] - order) * order_penalty <= eval_best:
                    A_best = A
                    B_best = B
                    C_best = C
                    D_best = D
                    order = A.shape[0]
                    eval_best = e_val

            if eval_best < eval_0:
                self.A = A_best
                self.B = B_best
                self.C = C_best
                self.D = D_best
                self.order = order
                self.eval_best = eval_best

            else:
                self.A = A
                self.B = B
                self.C = C
                self.D = D
                print('Solution Not Found')

        elif self.method == 'DMDc':
            pass

        elif self.method == 'N4SID':
            pass

        self.status = True

    def simulate(self, u, x0):
        A = self.A
        B = self.B
        C = self.C
        D = self.D

        steps = u.shape[0]

        xh = zeros((steps, A.shape[0]))
        yh = zeros((steps, C.shape[0]))
        xh[0] = x0

        for i in range(1, steps):
            xh[i] = A.dot(xh[i - 1]) + B.dot(u[i - 1])
            yh[i] = C.dot(xh[i - 1]) + D.dot(u[i - 1])

        return xh, yh

    def forecast(self, df: DataFrame = None):
        u_df = df[self.mv]

        u0 = u_df.to_numpy()

        u = u0 - self.u_ss

        x0 = zeros((self.A.shape[0],))

        x, ys = self.simulate(u, x0)

        ys += self.y_ss

        yp_df = DataFrame(ys, columns=self.label, index=df.index)
        yp_df = yp_df.shift(-1)
        yp_df.fillna(method='ffill', inplace=True)

        self.yp = yp_df

        return yp_df

# from pandas import read_csv
# import matplotlib.pyplot as plt
#
#
# x_df = read_csv('../../signal_df.csv')
# x_df.set_index('Time', drop=True, inplace=True)
# x_df = x_df.iloc[50:]
# ss = Subspace()
# ss.mv = ['F_cw']
# ss.cv = ['T1', 'T2', ]
# ss.identify(x_df)
# yp = ss.forecast(x_df)
# plt.plot(yp)
# plt.show()



from setuptools import setup, find_packages

with open(file='README.md', mode='r') as readme_handle:
    long_description = readme_handle.read()

setup(
    name='seeq_sysid',
    version='0.1',
    packages=find_packages(),
    url='https://github.com/BYU-PRISM/Seeq',
    license='MIT',
    author='Junho Park, Mohammad Reza Babaei',
    author_email='jnho.park@gmail.com, babaei_mr@outlook.com',
    description='Seeq System Identification Addon',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'ipyvuetify',
        'ipywidgets',
        'gekko',
        'numpy',
        'pandas',
        'plotly',
        'seeq',
        'tensorflow>=2.1',
        'keras-tuner==1.0.4'
    ],
    keywords='Seeq Prism System Identification ARX Subspace State-Space Neural-Network Addon Time-Series',
    include_package_data=True,
    package_data={'seeq_sysid': ['data/seeq_logo.png']},
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
    ]
)
